/* Copyright(C) 1997, Swedish Institute of Computer Science */

%%% Target library for arithmetic expressions:
%%%	'x+y=t'/3
%%%	'ax=t'/3
%%%	'ax+y=t'/4
%%%	'ax+by=t'/5
%%%	't=c'/2	
%%%	't=<c'/2
%%%	't\\=c'/2
%%%	't>=c'/2
%%%	't=u+c'/3
%%%	't=<u+c'/3
%%%	't\\=u+c'/3
%%%	't>=u+c'/3
%%%	't+u=c'/3
%%%	't+u=<c'/3
%%%	't+u\\=c'/3
%%%	't+u>=c'/3
%%%	't=x+y+c'/4
%%%	'x+y=u+c'/4
%%%	'x+y+z=c'/4

/***
't=c'(T,C) +:
	T in {C}.

't=<c'(T,C) +:
	T in inf..C.

't>=c'(T,C) +:
	T in C..sup.

't\\=c'(T,C) +:
	T in \{C}.

Type checking done at compile time.
***/

't=c'(T,C) :- propagate_interval(T, C, C).

't=<c'(T,C) :- propagate_interval(T, inf, C).

't>=c'(T,C) :- propagate_interval(T, C, sup).

't\\=c'(T,C) :-
	'$fd_range'(C, C, R),
	'$fd_dom_complement'(R, R1),
	prune_and_propagate(T, R1).


'ax=t'(A,X,T) +:
	X in min(T) /> A..max(T) /< A,
	T in min(X) *  A..max(X) *  A !.



'x+y=t'(X,Y,T) +:
	X in min(T) - max(Y)..max(T) - min(Y) !,
	Y in min(T) - max(X)..max(T) - min(X) !,
	T in min(X) + min(Y)..max(X) + max(Y) !.

% Now domain consistent!
't+u=c'(T,U,C) +:
	T in C - dom(U) !,
	U in C - dom(T) !.

% Now domain consistent!
'x+c=y'(X,C,Y) +:
	X in dom(Y) - C !,
	Y in dom(X) + C !.

% Now domain consistent!
't=u+c'(T,U,C) +:
	T in dom(U) + C !,
	U in dom(T) - C !.

't=<u+c'(T,U,C) +:
	T in inf..max(U)+C,
	U in min(T) - C..sup.

't\\=u+c'(T,U,C) +:
	T in \{U + C},
	U in \{T - C}.

't>=u+c'(T,U,C) +:
	T in min(U) + C..sup,
	U in inf..max(T) - C.

'ax+c=t'(A,X,C,Y) +:
	X in (min(Y) - C) /> A..(max(Y) - C) /< A,
	Y in  min(X)*A + C   .. max(X)*A + C !.

'ax+y=t'(A,X,Y,Z) +:
	X in (min(Z) - max(Y)) /> A..(max(Z) - min(Y)) /< A,
	Y in  min(Z) - max(X)*A	  .. max(Z) - min(X)*A !,
	Z in  min(X)*A + min(Y)	  .. max(X)*A + max(Y) !.

't+u=<c'(T,U,C) +:
	T in inf..C - min(U),
	U in inf..C - min(T).

't+u\\=c'(T,U,C) +:
	T in \{C - U},
	U in \{C - T}.

't+u>=c'(T,U,C) +:
	T in C - max(U)..sup,
	U in C - max(T)..sup.

% obsolete, backward compatibility?
'ax+by=t'(A,X,B,Y,Z) +:
	X in  (min(Z) - max(Y)*B) /> A.. (max(Z) - min(Y)*B) /< A,
	Y in  (min(Z) - max(X)*A) /> B.. (max(Z) - min(X)*A) /< B,
	Z in min(X)*A + min(Y)*B     ..max(X)*A + max(Y)*B !.

'x+y=u+c'(X,Y,U,C) +:
	X in min(U) - max(Y) + C..max(U) - min(Y) + C !,
	Y in min(U) - max(X) + C..max(U) - min(X) + C !,
	U in min(X) + min(Y) - C..max(X) + max(Y) - C !.

't=x+y+c'(T,X,Y,C) :-
	'x+y+c=z'(X,Y,C,T).

'x+y+c=z'(X,Y,C,Z) +:
	X in min(Z) - max(Y) - C..max(Z) - min(Y) - C !,
	Y in min(Z) - max(X) - C..max(Z) - min(X) - C !,
	Z in min(X) + min(Y) + C..max(X) + max(Y) + C !.

'ax+y+c=z'(A,X,Y,C,Z) +:
	X in  (min(Z) -	  max(Y) - C)/>A.. (max(Z) -   min(Y) - C)/<A,
	Y in   min(Z) - max(X)*A - C	..  max(Z) - min(X)*A - C !,
	Z in min(X)*A +	  min(Y) + C	..max(X)*A +   max(Y) + C !.

'x+y+z=t'(X,Y,Z,T) +:
	X in min(T) - max(Y) - max(Z)..max(T) - min(Y) - min(Z) !,
	Y in min(T) - max(X) - max(Z)..max(T) - min(X) - min(Z) !,
	Z in min(T) - max(X) - max(Y)..max(T) - min(X) - min(Y) !,
	T in min(X) + min(Y) + min(Z)..max(X) + max(Y) + max(Z) !.

'x+y+z=c'(X,Y,Z,C) +:
	X in C - max(Y) - max(Z)..C - min(Y) - min(Z) !,
	Y in C - max(X) - max(Z)..C - min(X) - min(Z) !,
	Z in C - max(X) - max(Y)..C - min(X) - min(Y) !.

'ax+y+z=t'(A,X,Y,Z,T) +:
	X in  (min(T) -	  max(Y) - max(Z)) /> A..
	      (max(T) -	  min(Y) - min(Z)) /< A,

	Y in   min(T) - max(X)*A - max(Z).. 
	       max(T) - min(X)*A - min(Z) !,

	Z in   min(T) - max(X)*A - max(Y)..
	       max(T) - min(X)*A - min(Y) !,

	T in min(X)*A +	  min(Y) + min(Z)..
	     max(X)*A +	  max(Y) + max(Z) !.


%%% Utilities for globals.

arg_mutable(X, Mut, _, _) :-
	var(X),
	propagate_interval(X, inf, sup),
	get_atts(X, fd_attribute(_,DomM,_)), !,
	Mut = DomM.
arg_mutable(X, DomM, _, _) :-
	integer(X),
	'$fd_range'(X, X, R), !,
	create_mutable(dom(R,X,X,1), DomM).
arg_mutable(X, _, Goal, ArgNo) :-
	fd_argument_error(Goal, ArgNo, X).

finite_arg_mutable(X, Mut, _, _) :-
	var(X),
	get_atts(X, fd_attribute(_,DomM,_)),
	get_mutable(Dom, DomM),
	arg(4, Dom, N),
	integer(N), !,
	Mut = DomM.
finite_arg_mutable(X, _, Goal, ArgNo) :-
	var(X),
	raise_exception(instantiation_error(Goal,ArgNo)).
finite_arg_mutable(X, DomM, _, _) :-
	integer(X),
	'$fd_range'(X, X, R), !,
	create_mutable(dom(R,X,X,1), DomM).
finite_arg_mutable(X, _, Goal, ArgNo) :-
	fd_argument_error(Goal, ArgNo, X).

must_be_dvar_list(List, Goal, ArgNo) :- var(List), !,
	fd_illarg(var, Goal, ArgNo, List).
must_be_dvar_list([], _, _) :- !.
must_be_dvar_list([Arg|List], Goal, ArgNo) :- !,
	(   var(Arg) -> propagate_interval(Arg, inf, sup)
	;   integer(Arg),
	    \+prolog:'$large_data'(0, Arg, _) -> true
	;   fd_argument_error(Goal, ArgNo, Arg)
	),
	must_be_dvar_list(List, Goal, ArgNo).
must_be_dvar_list(List, Goal, ArgNo) :-
	fd_illarg(domain(term,list), Goal, ArgNo, List).

/****************************************************************/
/* new all_different/[1,2], all_distinct/[1,2]                  */
/****************************************************************/

all_different(Xs) :-
	all_different(Xs, [], opt(value,false), all_different(Xs)).

all_different(Xs, Opt) :-
	all_different(Xs, Opt, opt(value,false), all_different(Xs,Opt)).

all_distinct(Xs) :-
	all_different(Xs, [], opt(domain,true), all_distinct(Xs)).

all_distinct(Xs, Opt) :-
	all_different(Xs, Opt, opt(domain,true), all_distinct(Xs,Opt)).

all_different(Xs, Options, Opt0, Goal) :-
	all_diff_options(Options, Opt0, Opt, Goal, 2),
	must_be_dvar_list(Xs, Goal, 1),
	length(Xs, N),
	Opt = opt(On,Complete),
	(   Complete==true
        ->  length(Xs, N),
	    all_dist_init(Xs, Vec, Mates, On, Goal, 1, Susp, []),
	    fd_global(all_distinct(Xs), f(N,Vec,Mates), Susp)
	;   all_diff_init(Xs, Vec, On, Goal, 1, Susp, []),
	    fd_global(all_different(Xs), Vec, Susp)
	).

all_diff_init([], [], _On, _Goal, _ArgNo) --> [].
all_diff_init([X|Xs], [X-M|Vec], On, Goal, ArgNo) -->
	on(On, X),
	{arg_mutable(X, M, Goal, ArgNo)},
	all_diff_init(Xs, Vec, On, Goal, ArgNo).

all_dist_init([], [], [], _On, _Goal, _ArgNo) --> [].
all_dist_init([X|Xs], [X-M|Vec], [Min|Mates], On, Goal, ArgNo) -->
	on(On, X),
	{arg_mutable(X, M, Goal, ArgNo)},
	{fd_min(X, Min)},
	all_dist_init(Xs, Vec, Mates, On, Goal, ArgNo).

on(_, X) --> {nonvar(X)}, !.
on(domain, X) --> [dom(X)].
on(range, X) --> [minmax(X)].
on(value, X) --> [val(X)].

all_diff_options(L, _, _, Goal, ArgNo) :- 
	var(L), !,
	fd_illarg(var, Goal, ArgNo).
all_diff_options([], Opt, Opt, _, _) :- !.
all_diff_options([X|L], Opt0, Opt, Goal, ArgNo) :- !,
	(   all_diff_option(X, Opt0, Opt1) -> true
        ;   fd_illarg(domain(term,all_diff_option), Goal, ArgNo, X)
        ),
	all_diff_options(L, Opt1, Opt, Goal, ArgNo).
all_diff_options(L, _, _, Goal, ArgNo) :- 
	fd_illarg(domain(term,list), Goal, ArgNo, L).

all_diff_option(on(On), opt(_,Complete), opt(On,Complete)) :-
	on(On, _, _, _).
all_diff_option(complete(Bool), opt(On,_), opt(On,Bool)) :-
	bool_option(Bool, _).


/****************************************************************/
/* element/3							*/
/****************************************************************/
% 
% A generalized version of element(X,L,Y):
% nonground L -> constructive disjunction style
% Propagator in C
% interval consistent in Y and L, domain consistent in X

element(X, L, Y) :-
	Goal = element(X,L,Y),
	length(L, N),
	X in 1..N,
	arg_mutable(X, XM, Goal, 1),
	arg_mutable(Y, YM, Goal, 3),
	element_cd_list(L, 0, LM, Susp, Goal, 2),
	fd_global(Goal, elt(N,X-XM,LM,Y-YM), [dom(X),minmax(Y)|Susp]).

element_cd_list([], _, [], [], _Goal, _ArgNo).
element_cd_list([V|Vs], I, [f(J,V,M)|Ms], [minmax(V)|Ps], Goal, ArgNo) :-
	J is I+1,
	arg_mutable(V, M, Goal, ArgNo),
	element_cd_list(Vs, J, Ms, Ps, Goal, ArgNo).

/****************************************************************/
/* circuit/[1,2]						*/
/****************************************************************/

circuit(Xs) :-
	circuit(Xs, _, circuit(Xs)).

circuit(Xs, Ys) :-
	circuit(Xs, Ys, circuit(Xs,Ys)).

circuit(Xs, Ys, Goal) :-
	must_be_dvar_list(Xs, Goal, 1),
	length(Xs, N),
	length(Ys, N),
	must_be_dvar_list(Ys, Goal, 2),
	(   N=:=1 -> Xs = [1], Ys = [1]
	;   '$fd_range'(1, N, Set),
	    domain(Xs, Set),
	    domain(Ys, Set),
	    not_self(Xs, 0),
	    not_self(Ys, 0),
	    assignment_state(Xs, Ys, 0, XVec, YVec, Mates, Heads,
			     domain, Goal, Susp, []),
	    fd_global(circuit(Xs,Ys), f(N,XVec,YVec,Mates,Heads), Susp)
	).

not_self([], _).
not_self([Y|Ys], I) :-
	J is I+1,
	't\\=c'(Y, J),
	not_self(Ys, J).

/****************************************************************/
/* relation/3							*/
/****************************************************************/
% 
% AC3-based implementation of relation/3
% 

relation(X, L0, Y) :-
	keysort(L0, L),				% must be keysorted
	map2list(L, L1, []),
	length(L1, N),
	Goal = relation(X,L0,Y),
	arg_mutable(X, XM, Goal, 1),
	arg_mutable(Y, YM, Goal, 3),
	fd_global(relation(X,L0,Y), f(N,X-XM,L1,Y-YM), [dom(X),dom(Y)]).

map2list([]) --> [].
map2list([X-Expr|Map]) -->
	{set_expression(Expr, Set),
	 fdset_to_list(Set, List, [])},
	tag_list(List, X),
	map2list(Map).

tag_list([], _) --> [].
tag_list([X|Xs], K) --> [K-X], tag_list(Xs, K).

/****************************************************************/
/* assignment/[2,3]						*/
/****************************************************************/

% assignment(X1...Xn, Y1...Yn) is true if 
% all Xi,Yi in 1..n and Xi=j iff Yj=i

assignment(Xs, Ys) :-
	assignment(Xs, Ys, [], opt(domain,true), assignment(Xs,Ys)).

assignment(Xs, Ys, Options) :-
	assignment(Xs, Ys, Options, opt(domain,true), assignment(Xs,Ys,Options)).

assignment(Xs, Ys, Options, Opt0, Goal) :-
	all_diff_options(Options, Opt0, Opt, Goal, 3),
	must_be_dvar_list(Xs, Goal, 1),
	length(Xs, N),
	length(Ys, N),
	must_be_dvar_list(Ys, Goal, 2),
	'$fd_range'(1, N, Set),
	domain(Xs, Set),
	domain(Ys, Set),
	Opt = opt(On,_Complete),
	assignment_state(Xs, Ys, 0, XVec, YVec, Mates, _,
			 On, Goal, Susp, []),
	fd_global(assignment(Xs,Ys), f(N,XVec,YVec,Mates), Susp).

assignment_state([], [], _, [], [], [], [], _On, _Goal) --> [].
assignment_state([X|Xs], [Y|Ys], I, [f(X,XM,J)|XVec], [f(Y,YM,J)|YVec], [Min|Mates], [J|Heads], On, Goal) -->
	on(On, X),
	on(On, Y),
	{arg_mutable(X, XM, Goal, 1),
	 arg_mutable(Y, YM, Goal, 2),
	 fd_min(X, Min),
	 J is I+1},
	assignment_state(Xs, Ys, J, XVec, YVec, Mates, Heads, On, Goal).

/****************************************************************/
/* serialized/2, cumulative/4 and friends			*/
/****************************************************************/

% serialized/[2,3], as of 3.8
serialized(Ss, Ds) :-
	serialized(Ss, Ds, []).

serialized(Ss, Ds, Options) :-
	Goal = serialized(Ss,Ds,Options),
	serialized_options(Options, opt(_,[],0), opt(R,Ps,Flags), Goal, 3),
	must_be_list_of_finite_dvar(Ss, Goal, 1),
	must_be_list_of_finite_dvar(Ds, Goal, 2),
	R = resource(Tasks,Diffs,Global),
	Rs = [1|Rs],
	cumulative(Ss, Ds, Rs, Ps, Tasks, Diffs, 1, Flags, Global, Goal).


% cumulative/[4,5] as of 3.8
cumulative(Ss, Ds, Rs, Limit) :-
	cumulative(Ss, Ds, Rs, Limit, []).

cumulative(Ss, Ds, Rs, Limit, Options) :-
	Goal = cumulative(Ss,Ds,Rs,Limit,Options),
	serialized_options(Options, opt(_,[],0), opt(R,Ps,Flags), Goal, 5),
	must_be_list_of_finite_dvar(Ss, Goal, 1),
	must_be_list_of_finite_dvar(Ds, Goal, 2),
	must_be_list_of_finite_dvar(Rs, Goal, 3),
	R = resource(Tasks,Diffs,Global),
	cumulative(Ss, Ds, Rs, Ps, Tasks, Diffs, Limit, Flags, Global, Goal).


serialized_options(L, _, _, Goal, ArgNo) :- 
	var(L), !,
	fd_illarg(var, Goal, ArgNo).
serialized_options([], Opt, Opt, _, _) :- !.
serialized_options([X|L], Opt0, Opt, Goal, ArgNo) :- !,
	(   serialized_option(X, Goal, Opt0, Opt1) -> true
        ;   fd_illarg(domain(term,serialized_option), Goal, ArgNo, X)
        ),
	serialized_options(L, Opt1, Opt, Goal, ArgNo).
serialized_options(L, _, _, Goal, ArgNo) :- 
	fd_illarg(domain(term,list), Goal, ArgNo, L).

% opt(Resource,Precedences,2'DCBA) where
% A = path_consistency
% B = static_sets
% C = edge_finder
% D = decomposition
serialized_option(-, _, _, _) :- !, fail.
serialized_option(precedences(Ps), _, opt(R,_,Flags), opt(R,Ps,Flags)).
serialized_option(resource(R), _, opt(_,Ps,Flags), opt(R,Ps,Flags)).
serialized_option(path_consistency(B), _, opt(R,Ps,Flags0), opt(R,Ps,Flags)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -2) \/ Value.
serialized_option(static_sets(B), _, opt(R,Ps,Flags0), opt(R,Ps,Flags)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -3) \/ (Value<<1).
serialized_option(edge_finder(B), _, opt(R,Ps,Flags0), opt(R,Ps,Flags)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -5) \/ (Value<<2).
serialized_option(decomposition(B), _, opt(R,Ps,Flags0), opt(R,Ps,Flags)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -9) \/ (Value<<3).

bool_option(-, _) :- !, fail.
bool_option(false, 0).
bool_option(true, 1).

/*
serialized_resource(Ss, Ds, resource(Tasks,Diffs,Global)) :-
	serialized(Ss, Ds, [], Tasks, Diffs, Global,
	           serialized_resource(Ss,Ds,resource(Tasks,Diffs,Global))).

serialized(Ss, Ds) :-
	serialized(Ss, Ds, [], _Tasks, _Diffs, _Global, serialized(Ss,Ds)).

serialized_precedence_resource(Ss, Ds, Ps, resource(Tasks,Diffs,Global)) :-
	serialized(Ss, Ds, Ps, Tasks, Diffs, Global,
	           serialized_precedence_resource(Ss,Ds,Ps,resource(Tasks,Diffs,Global))).

serialized_precedence(Ss, Ds, Ps) :-
	serialized(Ss, Ds, Ps, _Tasks, _Diffs, _Global,
		   serialized_precedence(Ss,Ds,Ps)).
*/

% guts

cumulative([], [], _Rs, _Ps, [], [], _, _, global, _Goal) :- !.
cumulative(Ss, Ds, Rs, Ps, Tasks, Diffs, Limit, Flags, Global, Goal) :-
	% inf_sup(Set1),
	% zero_sup(Set2),
	% must_be_dvar_list(Ss, Set1, Goal, 1),
	% must_be_dvar_list(Ds, Set2, Goal, 2),
	empty_assoc(H0),
	ti_delta_holes(Ps, H0, H),
	sort(Ps, Ps1),
	mktasks(Ss, Ds, Rs, Ps1, Tasks, 1, Goal), % 1-based numbering
	fd_max(Limit, LimitUB),
	ti_deltas(Tasks, 1, H, LimitUB, Goal, Diffs, []), % 1-based numbering
	task_order(Tasks, Diffs, Limit, Flags, Global).

mktasks([], _, _, _, [], _NT, _).
mktasks([S|Ss], [D|Ds], [R|Rs], Ps0, [task(S,SM,D,DM,R,RM,I,L3)|Ts], I, Goal) :-
	finite_arg_mutable(S, SM, Goal, 1),
	finite_arg_mutable(D, DM, Goal, 2),
	finite_arg_mutable(R, RM, Goal, 3),
	prec_keylist(Ps0, I, L1, Ps),		% key=distance, value=taskID
	keysort(L1, L2),
	reverse(L2, L3),			% by descending distance
	J is I+1,
	mktasks(Ss, Ds, Rs, Ps, Ts, J, Goal).

prec_keylist([D|Ps0], I, [Dis-J|L], Ps) :-
	D = d(I,J,Dis), !,
	prec_keylist(Ps0, I, L, Ps).
prec_keylist(Ps, _, [], Ps).

ti_delta_holes([], H, H).
ti_delta_holes([d(I,J,D)|Ds], H0, H) :-
	(   I<J ->
	    ti_one_less_pos(D, X),
	    ti_delta_hole(I-J, 1, X, H0, H1)
	;   ti_one_less_neg(D, X),
	    ti_delta_hole(J-I, X, -1, H0, H1)
	),
	ti_delta_holes(Ds, H1, H).

ti_delta_hole(_  , LB, UB, H0, H) :-
	integer(LB),
	integer(UB),
	LB > UB, !,
	H = H0.
ti_delta_hole(Key, LB, UB, H0, H) :-
	get_assoc(Key, H0, Hole0, H, Hole), !,
	fdset_parts(Part, LB, UB, []),
	'$fd_dom_union'(Hole0, Part, Hole).
ti_delta_hole(Key, LB, UB, H0, H) :-
	put_assoc(Key, H0, Hole, H), !,
	fdset_parts(Hole, LB, UB, []).

ti_one_less_pos(P, X) :- integer(P), !, X is P-1.
ti_one_less_pos(sup, sup).

ti_one_less_neg(P, X) :- integer(P), !, X is 1-P.
ti_one_less_neg(sup, inf).

ti_deltas([_], _, _H, _, _Goal) --> !.
ti_deltas([Taski|Tasks], I, H, Limit, Goal) -->
	{Taski = task(_,SM,_,DM,_,RM,_,_),
	 get_mutable(SD, SM),
	 SD = dom(_,Mini,Maxi,_),
	 get_mutable(DD, DM),
	 DD = dom(_,Duri,_,_),
	 get_mutable(RD, RM),
	 RD = dom(_,Resi,_,_),
	 MaxRes is Limit-Resi,
	 J is I+1},
	ti_deltas(Tasks, I, J, H, Mini, Maxi, Duri, MaxRes, Goal),
	ti_deltas(Tasks, J, H, Limit, Goal).

ti_deltas([], _, _, _H, _, _, _, _, _Goal) --> [].
ti_deltas([Taskj|Tasks], I, J, H, Mini, Maxi, Duri, MaxRes, Goal) -->
	[d(I,J,Mij)],
	{Taskj = task(_,SM,_,DM,_,RM,_,_),
	 get_mutable(SD, SM),
	 SD = dom(_,Minj,Maxj,_),
	 get_mutable(DD, DM),
	 DD = dom(_,Durj,_,_),
	 get_mutable(RD, RM),
	 RD = dom(_,Resj,_,_),
	 A is Minj-Maxi,
	 D is Maxj-Mini,
	 fdset_parts(Set0, A, D, []),
	 (   Resj>MaxRes ->
	     ti_constrain_dur(Duri, Durj, Set0, Set1)
	 ;   Set1 = Set0
	 ),
	 ti_constrain_precedence(I, J, H, Set1, Set),
	 Set \== [],
	 create_mutable(Set, Mij),
	 K is J+1},
	ti_deltas(Tasks, I, K, H, Mini, Maxi, Duri, MaxRes, Goal).

% TODO: For now, any task can be pre-empted by a zero-duration task.
ti_constrain_dur(0, _, Set, Set) :- !.
ti_constrain_dur(_, 0, Set, Set) :- !.
ti_constrain_dur(Duri, Durj, Set0, Set) :-
	Subl is 1-Durj,
	Subu is Duri-1,
	fdset_parts(Sub, Subl, Subu, []),
	'$fd_dom_subtract'(Set0, Sub, Set).

ti_constrain_precedence(I, J, H, Set0, Set) :-
	get_assoc(I-J, H, Hole), !,
	'$fd_dom_subtract'(Set0, Hole, Set).
ti_constrain_precedence(_, _, _, Set, Set).

get_precedence(Ps, I, J, Dur, Max) :-
	get_assoc(I-J, Ps, Prec),
	Prec > Dur, !,
	Max = Prec.
get_precedence(_, _, _, Dur, Dur).

task_order(Tasks, Diffs, Limit, Flags, Global) :-
	task_order_susp1(Tasks, Susp, [max(Limit)]),
	fd_global(task_order(Tasks,Diffs,Limit,Flags),
		  [chunk(Tasks,Diffs)], Susp, Global, clpfd).

task_order_susp1([]) --> [].
task_order_susp1([task(S,_,D,_,R,_,_,_)|L]) -->
	[minmax(S),min(D),min(R)],
	task_order_susp1(L).

'task_order solver'(Chunks0, Chunks, Limit, Flags, Actions) :-
	fd_max(Limit, LimitUB),
	(   Flags /\ 8 =:= 0 -> Chunks = Chunks0
        ;					% decompose the constraint
	    tasks_new_chunks(Chunks0, Chunks, [])
	),
        solve_task_chunks(Chunks, LimitUB, Flags, [exit], Actions).

tasks_new_chunks([]) --> [].
tasks_new_chunks([chunk(Tasks,Diffs)|Chunks]) -->
	{tag_by_est(Tasks, TasksByEST0),
	 keysort(TasksByEST0, TasksByEST),
	 tagged_task_ids(TasksByEST, IDs)},
	task_chunks(TasksByEST, Diffs, IDs, 1),
	tasks_new_chunks(Chunks).

tag_by_est([], []).
tag_by_est([Task|Tasks], [EST-Task|Tagged]) :-
	Task = task(_,SM,_,_,_,_,_,_),
	get_mutable(SDom, SM),
	domain_min(SDom, EST),
	tag_by_est(Tasks, Tagged).

task_chunks([], _, _, _) --> [].
task_chunks([EST-Task|Tasks], Diffs, IDs0, Flag) -->
	{IDs0 = [_|IDs]},
	{task_extent(Task, IDs, EST, LCT),
	 task_chunk(Tasks, EST, LCT, Chunk, Tasks1, Diffs, IDs, IDs1)},
	(   {Tasks1==[], Flag==1}
        ->  [chunk([Task|Chunk],Diffs)]	% no partitioning
	;   {Chunk==[]} -> []			% singleton chunk can be ignored
        ;   {task_ids([Task|Chunk], IdBag),
             sort(IdBag, IdSet),		% chunk was partitioned
	     filter_diffs(Diffs, IdSet, Diffs1)},
            [chunk([Task|Chunk],Diffs1)]
        ),
	task_chunks(Tasks1, Diffs, IDs1, 2).

task_chunk([], _, _, [], [], _Diffs, [], []).
task_chunk(Tasks0, _, LCT0, Chunk0, Tasks1, Diffs, IDs0, IDs1) :-
	Tasks0 = [EST-Task|Tasks],
	IDs0 = [_|IDs],
	(   LCT0==sup
        ->  Chunk0 = [Task|Chunk],
	    Tasks1 = [],
	    IDs1 = [],
	    values(Tasks, Chunk)
	;   LCT0 > EST
	->  Chunk0 = [Task|Chunk],
	    task_extent(Task, IDs, LCT0, LCT),
	    task_chunk(Tasks, EST, LCT, Chunk, Tasks1, Diffs, IDs, IDs1)
	;   Chunk0 = [],
	    Tasks1 = Tasks0,
	    IDs1 = IDs0
	).


task_extent(Task, IDs, LCT0, LCT) :-
	Task = task(_,SM,_,DM,_,_,_,Ps),
	get_mutable(SDom, SM),
	domain_max(SDom, LST),
	get_mutable(DDom, DM),
	domain_max(DDom, Dur0),
	max_prec(Ps, IDs, Dur1),
	(   integer(Dur1) -> LCT is max(LST+max(Dur0,Dur1),LCT0)
        ;   LCT = sup
        ).

max_prec([], _, 0).
max_prec([Dur1-J|Ps], IDs, Dur) :-
	(   memberchk(J, IDs) -> Dur = Dur1
        ;   max_prec(Ps, IDs, Dur)
        ).

solve_task_chunks([], _, _, Default, Default).
solve_task_chunks([chunk(Tasks,Diffs)|Chunks], LimitUB, Flags, _, Actions0) :-
	length(Tasks, NT),
	'$fd_task_order'(Tasks, Diffs, NT, LimitUB, Flags, Actions),
	(   Chunks==[] -> Actions0 = Actions
	;   append(Actions, Actions1, Actions0),
	    solve_task_chunks(Chunks, LimitUB, Flags, [], Actions1)
	).

task_ids([], []).
task_ids([task(_,_,_,_,_,_,I,_)|Tasks], [I|L]) :-
	task_ids(Tasks, L).

tagged_task_ids([], []).
tagged_task_ids([_-task(_,_,_,_,_,_,I,_)|Tasks], [I|L]) :-
	tagged_task_ids(Tasks, L).

filter_diffs([], _, []).
filter_diffs([D2|Ds1], Set, [D2|Ds2]) :-
	D2 = d(I,J,_),
	ord_subset([I,J], Set), !,
	filter_diffs(Ds1, Set, Ds2).
filter_diffs([_|Ds1], Set, Ds2) :-
	filter_diffs(Ds1, Set, Ds2).



/****************************************************************/
/* count/4							*/
/****************************************************************/

count(Value, Xs, Rel, Card) :-
	fdset_singleton(Set, Value),
	count_aux(Xs, Set, Bs),
	sum(Bs, Rel, Card).

count_aux([], _, []).
count_aux([X|Xs], Set, [B|Bs]) :-
	in_set_iff(X, Set, B),
	count_aux(Xs, Set, Bs).


/****************************************************************/
/* sum/3, scalar_product/4					*/
/****************************************************************/

sum(Xs, Rel, S) :-
	ones(Xs, Cs),
	scalar_product(Cs, Xs, Rel, S).

ones([], []).
ones([_|L1], [1|L2]) :- ones(L1, L2).


scalar_product(Cs, Xs, Rel, S) :-
	Goal = scalar_product(Cs,Xs,Rel,S),
	must_be_dvar_list(Xs, Goal, 2),
	must_be_dvar_list([S], Goal, 4),
	scalar_state(Cs, Xs, Rel, S, Vec, Sum, Susp),
	(   Rel==(#=)
	->  length(Vec, N),
	    sp_strength_reduce(N, Vec, Sum, Goal, Susp)
	;   fd_global(Goal, state(Vec,Sum), Susp)
        ).

%%% This is not an exhaustive list.

sp_strength_reduce(0, [], S, _Goal, _Susp) :- !, S = 0.
sp_strength_reduce(1, [f(C1,X1,_M1)], S, _Goal, _Susp) :- !,
	S mod C1 =:= 0,
	S1 is S//C1,
	't=c'(X1, S1).
sp_strength_reduce(2, [f(C1,X1,M1),f(C2,X2,M2)], S, Goal, Susp) :- !,
	sp_strength_reduce_2(C1, X1, M1, C2, X2, M2, S, Goal, Susp).
sp_strength_reduce(3, [f(C1,X1,M1),f(C2,X2,M2),f(C3,X3,M3)], S, Goal, Susp) :- !,
	sp_strength_reduce_3(C1, X1, M1, C2, X2, M2, C3, X3, M3, S, Goal, Susp).
sp_strength_reduce(_, Vec, Sum, Goal, Susp) :-
	fd_global(Goal, state(Vec,Sum), Susp).


sp_strength_reduce_2(1, X1, _M1, 1, X2, _M2, S, _Goal, _Susp) :- !,
	't+u=c'(X1, X2, S).
sp_strength_reduce_2(-1, X1, _M1, 1, X2, _M2, S, _Goal, _Susp) :- !,
	't=u+c'(X2, X1, S).
sp_strength_reduce_2(1, X1, _M1, -1, X2, _M2, S, _Goal, _Susp) :- !,
	't=u+c'(X1, X2, S).
sp_strength_reduce_2(-1, X1, _M1, -1, X2, _M2, S, _Goal, _Susp) :- !,
	S1 is -S, 
	't+u=c'(X1, X2, S1).
sp_strength_reduce_2(1, X1, _M1, C2, X2, _M2, S, _Goal, _Susp) :- C2>0, !,
	'ax+y=t'(C2, X2, X1, S).
sp_strength_reduce_2(1, X1, _M1, C2, X2, _M2, S, _Goal, _Susp) :- C2<0, !,
	C3 is -C2,
	'ax+y=t'(C3, X2, S, X1).
sp_strength_reduce_2(C1, X1, _M1, 1, X2, _M2, S, _Goal, _Susp) :- C1>0, !,
	'ax+y=t'(C1, X1, X2, S).
sp_strength_reduce_2(C1, X1, _M1, 1, X2, _M2, S, _Goal, _Susp) :- C1<0, !,
	C3 is -C1,
	'ax+y=t'(C3, X1, S, X2).
% sp_strength_reduce_2(C1, X1, _M1, C2, X2, _M2, S, _Goal, _Susp) :- C1>0, C2>0, !,
% 	S mod gcd(C1,C2) =:= 0,
% 	'ax+by=t'(C1, X1, C2, X2, S).
% sp_strength_reduce_2(C1, X1, _M1, C2, X2, _M2, S, _Goal, _Susp) :- C1<0, C2<0, !,
% 	C3 is -C1, C4 is -C2, S1 is -S,
% 	S mod gcd(C3,C4) =:= 0,
% 	'ax+by=t'(C3, X1, C4, X2, S1).
sp_strength_reduce_2(C1, X1, M1, C2, X2, M2, S, Goal, Susp) :-
	fd_global(Goal, state([f(C1,X1,M1),f(C2,X2,M2)],S), Susp).

sp_strength_reduce_3(1, X1, _M1, 1, X2, _M2, 1, X3, _M3, S, _Goal, _Susp) :- !,
	'x+y+z=c'(X1, X2, X3, S).
sp_strength_reduce_3(-1, X1, _M1, 1, X2, _M2, 1, X3, _M3, S, _Goal, _Susp) :- !,
	'x+y=u+c'(X2, X3, X1, S).
sp_strength_reduce_3(1, X1, _M1, -1, X2, _M2, 1, X3, _M3, S, _Goal, _Susp) :- !,
	'x+y=u+c'(X1, X3, X2, S).
sp_strength_reduce_3(-1, X1, _M1, -1, X2, _M2, 1, X3, _M3, S, _Goal, _Susp) :- !,
	't=x+y+c'(X3, X1, X2, S).
sp_strength_reduce_3(1, X1, _M1, 1, X2, _M2, -1, X3, _M3, S, _Goal, _Susp) :- !,
	'x+y=u+c'(X1, X2, X3, S).
sp_strength_reduce_3(-1, X1, _M1, 1, X2, _M2, -1, X3, _M3, S, _Goal, _Susp) :- !,
	't=x+y+c'(X2, X3, X1, S).
sp_strength_reduce_3(1, X1, _M1, -1, X2, _M2, -1, X3, _M3, S, _Goal, _Susp) :- !,
	't=x+y+c'(X1, X3, X2, S).
sp_strength_reduce_3(-1, X1, _M1, -1, X2, _M2, -1, X3, _M3, S, _Goal, _Susp) :- !,
	S1 is -S,
	'x+y+z=c'(X1, X2, X3, S1).
sp_strength_reduce_3(C1, X1, M1, C2, X2, M2, C3, X3, M3, S, Goal, Susp) :-
	fd_global(Goal, state([f(C1,X1,M1),f(C2,X2,M2),f(C3,X3,M3)],S), Susp).


'scalar_product solver'(#<, State0, State, Actions) :-
	'$fd_linear'(State0, State, inf, -1, Actions).
'scalar_product solver'(#=<, State0, State, Actions) :-
	'$fd_linear'(State0, State, inf, 0, Actions).
'scalar_product solver'(#>, State0, State, Actions) :-
	'$fd_linear'(State0, State, 1, sup, Actions).
'scalar_product solver'(#>=, State0, State, Actions) :-
	'$fd_linear'(State0, State, 0, sup, Actions).
'scalar_product solver'(#=, State0, State, Actions) :-
	'$fd_linear'(State0, State, 0, 0, Actions).
'scalar_product solver'(#\=, State0, State, Actions) :-
	'$fd_linear'(State0, State, inf, sup, Actions).

scalar_state(Cs, Xs, Rel, S, Vec, Sum, Susp) :-
	integer(S), !,
	scalar_vector(Cs, Xs, Rel, S, Vec, Sum, Susp).
scalar_state(Cs, Xs, Rel, S, Vec, Sum, Susp) :-
	scalar_vector([-1|Cs], [S|Xs], Rel, 0, Vec, Sum, Susp).

scalar_vector(Cs, Xs, Rel, S, Vec, Sum, Susp) :-
	scalar_pairs(Xs, Cs, L1),
	keysort(L1, L2),
	keyfuse(L2, L3),
	scalar_vector(L3, Rel, S, Vec, Sum, Susp).

scalar_pairs([], [], []).
scalar_pairs([X|Xs], [C|Cs], [X-C|XCs]) :- scalar_pairs(Xs, Cs, XCs).

scalar_vector([], _, Sum, [], Sum, []).
scalar_vector([X-C|XCs], Rel, S0, [f(C,X,M)|Vec], S, [Sus|Susp]) :- 
	var(X), C=\=0, !,
	get_fd_domain_mutable(X, M),
	susp_type(Rel, C, X, Sus),
	scalar_vector(XCs, Rel, S0, Vec, S, Susp).
scalar_vector([X-C|XCs], Rel, S0, Vec, S, Susp) :-
	C=\=0, !,
	S1 is S0-C*X,
	scalar_vector(XCs, Rel, S1, Vec, S, Susp).
scalar_vector([_|XCs], Rel, S0, Vec, S, Susp) :-
	scalar_vector(XCs, Rel, S0, Vec, S, Susp).

susp_type(#<, C, X, Type) :- susp_type(#=<, C, X, Type).
susp_type(#=<, C, X, Type) :- C>0 -> Type = min(X); Type = max(X).
susp_type(#>, C, X, Type) :- susp_type(#>=, C, X, Type).
susp_type(#>=, C, X, Type) :- C>0 -> Type = max(X); Type = min(X).
susp_type(#=, _, X, minmax(X)).
susp_type(#\=, _, X, val(X)).

/****************************************************************/
/* 'x*x=y'/2							*/
/****************************************************************/

'x*x=y'(X, Y) :-
	integer(X), !,
	Y is X*X.
'x*x=y'(X, Y) :-
	integer(Y), !,
	(   Y =:= 0 -> '$fd_range'(0, 0, R)
	;   Y >= 0,
	    X1 is integer(sqrt(Y)),
	    X1*X1 =:= Y,
	    X0 is -X1,
	    '$fd_range'(X0, X0, R0),
	    '$fd_dom_insert'(R0, X1, R)
	),
	prune_and_propagate(X, R).
'x*x=y'(X, Y) :-
	Goal = 'x*x=y'(X,Y),
	arg_mutable(X, XM, Goal, 1),
	arg_mutable(Y, YM, Goal, 2),
	Y in 0..sup,
	Susp = [minmax(X),minmax(Y)],
	fd_global(Goal, state(XM,YM), Susp).

'x*x=y solver'(X, Y, State, State, Actions) :-
	State = state(XM,YM),
	'$fd_square'(X, XM, Y, YM, Actions).

/****************************************************************/
/* 'x*y=z'/3							*/
/****************************************************************/

'x*y=z'(X, Y, Z) :-
	integer(X), !,
	(   X=:=0 -> Z = 0
	;   X > 0 -> 'ax=t'(X, Y, Z)
	;   X1 is -X, '-ax=t'(X1, Y, Z)
	).
'x*y=z'(X, Y, Z) :-
	integer(Y), !,
	'x*y=z'(Y, X, Z).
'x*y=z'(X, Y, Z) :-
	Goal = 'x*y=z'(X,Y,Z),
	arg_mutable(X, XM, Goal, 1),
	arg_mutable(Y, YM, Goal, 2),
	arg_mutable(Z, ZM, Goal, 3),
	Susp = [minmax(X),minmax(Y),minmax(Z)],
	fd_global(Goal, state(XM,YM,ZM), Susp).

'-ax=t'(A,X,T) +:
	X in -max(T) /< A .. -min(T) /> A,
	T in -max(X) *	A .. -min(X) *	A !.

'x*y=z solver'(X, Y, Z, State, State, Actions) :-
	State = state(XM,YM,ZM),
	'$fd_product'(X, XM, Y, YM, Z, ZM, Actions).

/****************************************************************/
/* 'x/y=z'/3							*/
/****************************************************************/

:- block 'x/y=z'(?, -, ?).
'x/y=z'(X, Y, Z) :-
	integer(X), !,
	Y =\= 0,
	Z is X//Y.
'x/y=z'(X, Y, Z) :-
	Y>0,
	fd_min(X, Xa), integer(Xa), Xa>=0, !,
	'x/k=z'(X, Y, Z).
'x/y=z'(X, Y, Z) :-
	Goal = 'x/y=z'(X,Y,Z),
	Y =\= 0,
	arg_mutable(X, XM, Goal, 1),
	arg_mutable(Z, ZM, Goal, 3),
	Susp = [minmax(X),minmax(Z)],
	fd_global(Goal, state(XM,ZM), Susp). 

'x/k=z'(X, K, Z) +:
	Z in min(X)/<K .. max(X)/<K,
	X in min(Z)*K  .. (max(Z)+1)*K - 1.

'x/y=z solver'(X, Y, Z, State, State, Actions) :-
	State = state(XM,ZM),
	'$fd_quotient'(X, XM, Y, Z, ZM, Actions).

/****************************************************************/
/* 'x mod y=z'/3						*/
/****************************************************************/

:- block 'x mod y=z'(?, -, ?).
'x mod y=z'(X, Y, Z) :-
	integer(X), !,
	Y =\= 0,
	Z is X mod Y.
'x mod y=z'(X, Y, Z) :-
	Goal = 'x mod y=z'(X,Y,Z),
	Y =\= 0,
	arg_mutable(X, XM, Goal, 1),
	arg_mutable(Z, ZM, Goal, 3),
	B is abs(Y)-1,
	A is -B,
	Z in A..B,
	Susp = [minmax(X),minmax(Z)],
	fd_global(Goal, state(XM,ZM), Susp).

'x mod y=z solver'(X, Y, Z, State, State, Actions) :-
	State = state(XM,ZM),
	'$fd_modulo'(X, XM, Y, Z, ZM, Actions).

/****************************************************************/
/* min/3, max/3 etc						*/
/****************************************************************/

'min(x,y)=z'(X, Y, Z) :-
	X #>= Z,
	Y #>= Z,
	'oneof(x,y)=z'(X, Y, Z).

'max(x,y)=z'(X, Y, Z) :-
	X #=< Z,
	Y #=< Z,
	'oneof(x,y)=z'(X, Y, Z).

'oneof(x,y)=z'(X, Y, Z) +:
	X in ((dom(Y)/\dom(Z)) ? (inf..sup)) \/ dom(Z) !,
	Y in ((dom(X)/\dom(Z)) ? (inf..sup)) \/ dom(Z) !,
	Z in (dom(X)\/dom(Y)) !.

'|x|=y'(X,Y) :-
	't>=c'(Y, 0),				% do this outside the loop
	'|x|=y 1'(X,Y).

'|x|=y 1'(X,Y) +:
	X in dom(Y) \/ (0-dom(Y)) !,
	Y in dom(X) \/ (0-dom(X)) !.

%%% Support for reified constraints (domain reasoning for now)

% Now domain consistent!
'x=y'(X,Y) +:
	X in dom(Y) !,
	Y in dom(X) !.
'x=y'(X,Y) -:
	X in \{Y},
	Y in \{X}.
% 'x=y'(X,Y) +?				% covered by next rule
%	X in {Y}.
'x=y'(X,Y) -?
	X in \dom(Y).

% Now domain consistent!
% just the negation of the above
'x\\=y'(X,Y) -:
	X in dom(Y) !,
	Y in dom(X) !.
'x\\=y'(X,Y) +:
	X in \{Y},
	Y in \{X}.
% 'x\\=y'(X,Y) -?			% covered by next rule
%	X in {Y}.
'x\\=y'(X,Y) +?
	X in \dom(Y).

'x=<y'(X,Y) +:
	X in inf..max(Y),
	Y in min(X)..sup.
'x=<y'(X,Y) -:
	X in (min(Y)+1)..sup,
	Y in inf..(max(X)-1).
'x=<y'(X,Y) +?
	X in inf..min(Y).
'x=<y'(X,Y) -?				% NOT covered by prev rule
	X in (max(Y)+1)..sup.

%%% Reified constraints.

iff_aux(Constraint0, B) :-
	prolog:get_module(Constraint0, Constraint, Module),
	iff_aux(Constraint, Module, B).

iff_aux(X in Expr, _, B) :- !,
	fd_goal_expansion(X in Expr #<=> B, clpfd, Goal),
	call(Goal).
iff_aux(Constraint, _, B) :-
	fd_expandable(Constraint, _, _, _), !,
	fd_goal_expansion(Constraint #<=> B, clpfd, Goal),
	call(Goal).
iff_aux(Constraint, Module, B) :-
	'$fd_find_definition'(Constraint, Module, DefPtr),
	(   DefPtr =\= 0
	->  check_arguments(Constraint, Domv),
	    iff_aux(DefPtr, Constraint, B, Domv)
	;   functor(Constraint, Name, Arity),
	    fd_illarg(existence(constraint,Module:Name/Arity,0), Constraint, 0)
	).

iff_aux(Def, Constraint, B, Domv) :-
	var(B), !,
	B in 0..1,
	enqueue_ix(Def, 2, Constraint, A, B, Domv, Q0, Q1),
	enqueue_ix(Def, 3, Constraint, A, B, Domv, Q1, Q),
	'$fd_get_indexicals'(Def, 0, Constraint, Ent, _, Domv, Ix0s),
	link_iff(Ix0s, B, 0, A),
	'$fd_get_indexicals'(Def, 1, Constraint, Ent, _, Domv, Ix1s),
	link_iff(Ix1s, B, 1, A),
	evaluate(Q0, Q, G, G).
iff_aux(Def, Constraint, B, Domv) :-
	integer(B),
	B /\ -2 =:= 0,
	enqueue_ix(Def, B, Constraint, _, _, Domv, Q0, Q),
	evaluate(Q0, Q, G, G).


link_iff([], _, _, _).
link_iff([Ix|Ixs], B, Key, A) :-
	'$fd_mon_vars'(Ix, Vars),
	'$fd_link'([val(B)], Vars-iff(Ix,B,Key,A)),
	link_iff(Ixs, B, Key, A).


enqueue_ix(DefPtr, B, Constraint, A, ZeroOne, Domv) -->
	{'$fd_get_indexicals'(DefPtr, B, Constraint, A, ZeroOne, Domv, Ixs)},
	enqueue_first(Ixs).


in_aux_rt(X, Expr) :-
	ground(Expr),
	set_expression(Expr, Set), !,
	prune_and_propagate(X, Set).
in_aux_rt(X, Expr) :-
	ill_formed_constraint(X in Expr), fail.

in_aux_rt(X, Expr, B) :-
	ground(Expr),
	set_expression(Expr, Set), !,
	in_set_iff(X, Set, B).
in_aux_rt(X, Expr, B) :-
	ill_formed_constraint(X in Expr #<=> B), fail.


in_set_aux_rt(X, Set) :-
	'$fd_size'(Set, _), !,
	prune_and_propagate(X, Set).
in_set_aux_rt(X, Set) :-
	ill_formed_constraint(X in_set Set), fail.

in_set_aux_rt(X, Set, B) :-
	'$fd_size'(Set, _), !,
	in_set_iff(X, Set, B).
in_set_aux_rt(X, Set, B) :-
	ill_formed_constraint(X in_set Set #<=> B), fail.

in_set_iff(X, Set, B) :-
	Goal = in_set_iff(X,Set,B),
	B in 0..1,
	arg_mutable(X, XMut, Goal, 1),
	arg_mutable(B, BMut, Goal, 3),
	fd_global(Goal, state(X,XMut,Set,B,BMut), [dom(X),dom(B)]).


/****************************************************************/
/* bool/4							*/
/****************************************************************/

bool(Fun, X, Y, Z) :-
	Goal = bool(Fun,X,Y,Z),
	arg_mutable(X, XMut, Goal, 1),
	arg_mutable(Y, YMut, Goal, 2),
	arg_mutable(Z, ZMut, Goal, 3),
	propagate_interval(X, 0, 1),
	propagate_interval(Y, 0, 1),
	propagate_interval(Z, 0, 1),
	fd_global(Goal, state(X,XMut,Y,YMut,Z,ZMut), [dom(X),dom(Y),dom(Z)]).

'bool solver'(F, State, State, Actions) :-
	State = state(X,XMut,Y,YMut,Z,ZMut),
	'$fd_bool'(F, X, XMut, Y, YMut, Z, ZMut, Actions).

% FOR BACKWARD COMPATIBILITY!

'\\p'(P, B) :- bool(6, P, B, 1).

'p/\\q'(P, Q, B) :- bool(0, P, Q, B).

'p\\q'(P, Q, B) :- bool(6, P, Q, B).

'p\\/q'(P, Q, B) :- bool(3, P, Q, B).

'p=>q'(P, Q, B) :- bool(4, Q, P, B).

'p<=>q'(P, Q, B) :- bool(7, P, Q, B).

/****************************************************************/
/* domain/3							*/
/****************************************************************/

domain(Vars, Min, Max) :-
	set_expression(Min..Max, Set), !,
	domain(Vars, Set).
domain(Vars, Min, Max) :-
	fd_illarg(domain(term,set_expression), domain(Vars,Min,Max), 0, Min..Max).

domain(Vars, Set) :-
	domain(Vars, Set, Q0, Q, G0, G),
	evaluate(Q0, Q, G0, G).

domain([], _, Q, Q, G, G).
domain([X|Xs], Set, Q0, Q, G0, G) :-
	'$fd_in_set'(X, Set, Q0, Q1, G0, G1),
	domain(Xs, Set, Q1, Q, G1, G).

%%% predicates corresponding to macro-expanded constraints

X in Expr :-
	in_aux_rt(X, Expr).

X in_set Set :-
	in_set_aux_rt(X, Set).

X #= Y :-
	fd_goal_expansion(X #= Y, clpfd, Goal),
	Goal.

X #\= Y :-
	fd_goal_expansion(X #\= Y, clpfd, Goal),
	Goal.

X #< Y :-
	fd_goal_expansion(X #< Y, clpfd, Goal),
	Goal.

X #=< Y :-
	fd_goal_expansion(X #=< Y, clpfd, Goal),
	Goal.

X #> Y :-
	fd_goal_expansion(X #> Y, clpfd, Goal),
	Goal.

X #>= Y :-
	fd_goal_expansion(X #>= Y, clpfd, Goal),
	Goal.

:- meta_predicate #\(:).
#\ Q :-
	fd_goal_expansion(#\ Q, clpfd, Goal),
	Goal.

:- meta_predicate #/\(:,:).
P #/\ Q :-
	fd_goal_expansion(P #/\ Q, clpfd, Goal),
	Goal.

:- meta_predicate #\(:,:).
P #\ Q :-
	fd_goal_expansion(P #\ Q, clpfd, Goal),
	Goal.

:- meta_predicate #\/(:,:).
P #\/ Q :-
	fd_goal_expansion(P #\/ Q, clpfd, Goal),
	Goal.

:- meta_predicate #=>(:,:).
P #=> Q :-
	fd_goal_expansion(P #=> Q, clpfd, Goal),
	Goal.

:- meta_predicate #<=(:,:).
P #<= Q :-
	fd_goal_expansion(P #<= Q, clpfd, Goal),
	Goal.

:- meta_predicate #<=>(:,:).
P #<=> Q :-
	fd_goal_expansion(P #<=> Q, clpfd, Goal),
	Goal.

/****************************************************************/
/* disjoint1/2, disjoint2/2                  			*/
/****************************************************************/

disjoint1(Items) :-
	disjoint1(Items, []).

disjoint1(Items, Options) :-
	Goal = disjoint1(Items,Options),
	disjoint1_options(Options, opt(0,inf,sup,[]), Opt, Goal, 2),
	(   Opt = opt(Flags,Min,B,_),
	    Flags /\ 2 =:= 2
	->  Max is B-1
	;   Min = inf, Max = sup
	),
	mkitems(Items, Items2, Min, Max, Goal, Susp, []),
	fd_global(Goal, state([Items2],Opt), Susp).


disjoint1_options(L, _, _, Goal, ArgNo) :- 
	var(L), !,
	fd_illarg(var, Goal, ArgNo).
disjoint1_options([], Opt, Opt, _, _) :- !.
disjoint1_options([X|L], Opt0, Opt, Goal, ArgNo) :- !,
	(   disjoint1_option(X, Goal, Opt0, Opt1) -> true
        ;   fd_illarg(domain(term,disjoint1_option), Goal, ArgNo, X)
        ),
	disjoint1_options(L, Opt1, Opt, Goal, ArgNo).
disjoint1_options(L, _, _, Goal, ArgNo) :- 
	fd_illarg(domain(term,list), Goal, ArgNo, L).

% opt(2'AMXD,Min,Max,Margins) where
% A = global
% M = Margins \== []
% X = wrap-around
% D = decomposition
% Min..Max is the interval subject to wrap-around
% Margins = list of margin(Type1,Type2,Diff) = list of extra margins
disjoint1_option(-, _, _, _) :- !, fail.
disjoint1_option(decomposition(B), _, opt(Flags0,Min,Max,Ms), opt(Flags,Min,Max,Ms)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -2) \/ Value.
disjoint1_option(global(B), _, opt(Flags0,Min,Max,Ms), opt(Flags,Min,Max,Ms)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -9) \/ (Value<<3).
disjoint1_option(wrap(Min,Max), _, opt(Flags0,_,_,Ms), opt(Flags,Min,Max,Ms)) :-
	(   Min==inf, Max==sup -> Flags is Flags0 /\ -3
	;   integer(Min),
	    integer(Max),
	    Min < Max,
	    Flags is (Flags0 /\ -3) \/ 2
	).
disjoint1_option(margin(T1,T2,D), _, opt(Flags0,Min,Max,Tail), opt(Flags,Min,Max,[margin(T1,T2,D)|Tail])) :-
	Flags is (Flags0 /\ -5) \/ 4.


mkitems([], [], _, _, _) --> [].
mkitems([X|Xs], [at(item(S,SM,DM,Type),V1,V1)|Items], Min, Max, Goal) --> [min(S),max(S),min(D)],
	{arg(1,X,S), arg(2,X,D)},
	{arg(3,X,Type) -> true; Type=0},
	{propagate_interval(S, Min, Max)},
	{finite_arg_mutable(S, SM, Goal, 1)},
	{finite_arg_mutable(D, DM, Goal, 1)},
	{fd_min(S, V0), V1 is V0-1}, % outside the domain
	mkitems(Xs, Items, Min, Max, Goal).

'disjoint1 solver'(state(Chunks0,Opt), state(Chunks,Opt), Actions) :-
	Opt = opt(Flags,_,_,_),
	(   Flags/\1 =:= 1 ->
	    disjoint1_all(Chunks0, Opt, [], Chunks, [], Actions)
	;   Chunks0 = [Chunk],
	    length(Chunk, N),
	    '$fd_disjoint1'(N, Opt, Chunk, Chunks, Actions)
	).

disjoint1_all([], _, NewChunks, NewChunks, Actions, Actions).
disjoint1_all([Chunk|Chunks], Opt, NewChunks0, NewChunks, Actions0, Actions) :-
	length(Chunk, N),
	'$fd_disjoint1'(N, Opt, Chunk, Chs, Acts),
	append(Chs, NewChunks0, NewChunks1),
	append_no_exit(Acts, Actions0, Actions1),
	disjoint1_all(Chunks, Opt, NewChunks1, NewChunks, Actions1, Actions).



% 2D version
disjoint2(Items) :-
	disjoint2(Items, []).

disjoint2(Items, Options) :-
	Goal = disjoint2(Items,Options),
	disjoint2_options(Options, opt(0,inf,sup,inf,sup,[]), Opt, Goal, 2),
	Opt = opt(Flags,Min1,B1,Min2,B2,_),
	(   Flags /\ 2 =:= 2
	->  Max1 is B1-1
	;   Min1 = inf, Max1 = sup
	),
	(   Flags /\ 4 =:= 4
	->  Max2 is B2-1
	;   Min2 = inf, Max2 = sup
	),
	mkitems(Items, Items2, Min1, Max1, Min2, Max2, Goal, Susp, []),
	fd_global(Goal, state([Items2],Opt), Susp).


disjoint2_options(L, _, _, Goal, ArgNo) :- 
	var(L), !,
	fd_illarg(var, Goal, ArgNo).
disjoint2_options([], Opt, Opt, _, _) :- !.
disjoint2_options([X|L], Opt0, Opt, Goal, ArgNo) :- !,
	(   disjoint2_option(X, Goal, Opt0, Opt1) -> true
        ;   fd_illarg(domain(term,disjoint2_option), Goal, ArgNo, X)
        ),
	disjoint2_options(L, Opt1, Opt, Goal, ArgNo).
disjoint2_options(L, _, _, Goal, ArgNo) :- 
	fd_illarg(domain(term,list), Goal, ArgNo, L).

% opt(2'AMYXD,Min1,Max1,Min2,Max2,Margins) where
% A = global
% M = Margins \== []
% Y = wrap-around in Y dim
% X = wrap-around in X dim
% D = decomposition
% Min..Max is the interval subject to wrap-around
% Margins = list of margin(Type1,Type2,Diff1,Diff2) = list of extra margins
disjoint2_option(-, _, _, _) :- !, fail.
disjoint2_option(decomposition(B), _, 
		 opt(Flags0,Min1,Max1,Min2,Max2,Ms), 
		 opt(Flags,Min1,Max1,Min2,Max2,Ms)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -2) \/ Value.
disjoint2_option(global(B), _, 
		 opt(Flags0,Min1,Max1,Min2,Max2,Ms), 
		 opt(Flags,Min1,Max1,Min2,Max2,Ms)) :-
	bool_option(B, Value),
	Flags is (Flags0 /\ -17) \/ (Value<<4).
disjoint2_option(wrap(Min1,Max1,Min2,Max2), _, 
		 opt(Flags0,_,_,_,_,Ms), 
		 opt(Flags,Min1,Max1,Min2,Max2,Ms)) :-
	(   Min1==inf, Max1==sup -> Flags1 is (Flags0 /\ -3)
	;   integer(Min1),
	    integer(Max1),
	    Min1 < Max1,
	    Flags1 is (Flags0 /\ -3) \/ 2
	),
	(   Min2==inf, Max2==sup -> Flags is (Flags1 /\ -5)
	;   integer(Min2),
	    integer(Max2),
	    Min2 < Max2,
	    Flags is (Flags1 /\ -5) \/ 4
	).
disjoint2_option(margin(T1,T2,D1,D2), _, 
		 opt(Flags0,Min1,Max1,Min2,Max2,Tail), 
		 opt(Flags,Min1,Max1,Min2,Max2,[margin(T1,T2,D1,D2)|Tail])) :-
	Flags is (Flags0 /\ -9) \/ 8.


% S1,S2 - start variables
% SM1,SM2 - start domain mutables
% D1,D2 - durations
% V1,V2,W1,W2 - witnesses
% Type - type of object (optional)
mkitems([], [], _, _, _, _, _) --> [].
mkitems([X|Xs], [at(item(S1,SM1,DM1,S2,SM2,DM2,Type),
		    V1,W1,V1,W1,V2,W2,V2,W2)|Items],
	Min1, Max1, Min2, Max2, Goal) -->
	[min(S1),max(S1),min(D1),min(S2),max(S2),min(D2)],
	{arg(1,X,S1), arg(2,X,D1), arg(3,X,S2), arg(4,X,D2)},
	{arg(5,X,Type) -> true; Type=0},
	{propagate_interval(S1, Min1, Max1)},
	{finite_arg_mutable(S1, SM1, Goal, 1)},
	{finite_arg_mutable(D1, DM1, Goal, 1)},
	{propagate_interval(S2, Min2, Max2)},
	{finite_arg_mutable(S2, SM2, Goal, 1)},
	{finite_arg_mutable(D2, DM2, Goal, 1)},
	{fd_min(S2, W1), fd_min(S1, W2)}, % TODO: randomize
	{V2 is W1-1, V1 is W2-1}, % outside the domains
	mkitems(Xs, Items, Min1, Max1, Min2, Max2, Goal).

'disjoint2 solver'(state(Chunks0,Opt), state(Chunks,Opt), Actions) :-
	Opt = opt(Flags,_,_,_,_,_),
	(   Flags/\1 =:= 1 ->
	    disjoint2_all(Chunks0, Opt, [], Chunks, [], Actions)
	;   Chunks0 = [Chunk],
	    length(Chunk, N),
	    '$fd_disjoint2'(N, Opt, Chunk, Chunks, Actions)
	).

disjoint2_all([], _, NewChunks, NewChunks, Actions, Actions).
disjoint2_all([Chunk|Chunks], Opt, NewChunks0, NewChunks, Actions0, Actions) :-
	length(Chunk, N),
	'$fd_disjoint2'(N, Opt, Chunk, Chs, Acts),
	append(Chs, NewChunks0, NewChunks1),
	append_no_exit(Acts, Actions0, Actions1),
	disjoint2_all(Chunks, Opt, NewChunks1, NewChunks, Actions1, Actions).

% the whole is only entailed if ALL chunks are entailed
% for now, simply disable entailment
append_no_exit([], List, List).
append_no_exit([exit|Tail], List, Rest) :- !,
	append_no_exit(Tail, List, Rest).
append_no_exit([Head|Tail], List, [Head|Rest]) :- 
	append_no_exit(Tail, List, Rest).


